# Auto-generated file, do not edit.
##
# version.py: Specifies the version of the qsharp package.
##
# Copyright (c) Microsoft Corporation.
# Licensed under the MIT License.
##
__version__ = "0.20.2110.171562b1"
is_conda = False
_user_agent_extra = "[0.20.2110.171562b1]"
